/* vi: set sw=4 ts=4: */
/*
 * CCE - Console Chinese Environment -
 * Copyright (C) 1998-2003 Rui He (rhe@3eti.com)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 */

#ifndef	__CCE_FB_H__
#define	__CCE_FB_H__

#if defined(linux)
#include    <linux/fb.h>
#include    <linux/types.h>
#elif defined(__FreeBSD__)
#include	<sys/fbio.h>
#include	<sys/consio.h>

extern int GetOrigFreeBSDVideoMode(void);
extern int GetFreeBSDVideoMode(int width, int height, int depth, video_info_t *vmode);
extern int GetBestFreeBSDConsVidMode(video_info_t *vmode);
extern int ConfigFreeBSDFrameBuffer(const char *confstr);

#endif

#if defined(linux)

#define DEFAULT_FB  		"/dev/fb"
#define DEFAULT_FB_DEVFS	"/dev/fb/0"

#elif defined(__FreeBSD__)
extern int fbwidth, fbheight, fbdepth;
#endif

extern void FBOpClear(int y, int height, int color);
extern void FBOpDraw(u_char *code, u_char fc, u_char bc, int width, int height);
extern void FBOpCursor(CursorInfo *ci, int xpos, int ypos);

#ifdef SUPPORT_FREETYPE_FONT
#include    "ftfont.h"
extern void FBDrawFreeTypeChar(int xpos, int ypos, FT_GlyphSlot glyph, int fc, int bc, int width, int height);
#endif

typedef union
{
    u_short cfb16[16];
    u_long  cfb24[16];
    u_long  cfb32[16];
} FBCMAP;

extern FBCMAP fb_cmap;

#ifdef __Darwin__
#define DARWIN_COLOR_MASK 0x07
extern u_char DarwinColorMap8[8];   /* Map colors from 0-7 to 0x00-0xFF, for 8bpp display */
#endif

#endif  /* __CCE_FB_H__ */
